@@ -1,12 +1,13 @@
-use crate::format::{self, Format};
-use crate::items::expressions::{Body, Expr};
-use crate::items::generics::{Clauses, Either, Maybe, NonEmptyItems, WithArrow, WithGuard};
+use crate::format::{Format, Formatter, Indent, Newline};
+use crate::items::components::{Clauses, Either, Maybe, NonEmptyItems, WithArrow, WithGuard};
+use crate::items::expressions::components::Body;
 use crate::items::keywords::{
     AfterKeyword, BeginKeyword, CaseKeyword, CatchKeyword, EndKeyword, IfKeyword, OfKeyword,
     ReceiveKeyword, TryKeyword,
 };
 use crate::items::symbols::{ColonSymbol, CommaSymbol, SemicolonSymbol};
 use crate::items::tokens::{AtomToken, VariableToken};
+use crate::items::Expr;
 use crate::parse::Parse;
 use crate::span::Span;
 
@@ -22,36 +23,36 @@ pub enum BlockExpr {
 
 /// `case` [Expr] `of` (`$CLAUSE` `;`?)+ `end`
 ///
-/// - $CLAUSE: `$PATTERN` (`when` `$GUARD`)? `->` `$BODY`
+/// - $CLAUSE: `$PATTERN` (`when` `$GUARD`)? `->` [Body]
 /// - $PATTERN: [Expr]
 /// - $GUARD: ([Expr] (`,` | `;`)?)+
-/// - $BODY: ([Expr] `,`?)+
 #[derive(Debug, Clone, Span, Parse)]
 pub struct CaseExpr {
     case: CaseKeyword,
     value: Expr,
     of: OfKeyword,
     clauses: Block<Clauses<CaseClause>>,
-    end: EndKeyword,
+    end: End,
 }
 
 impl Format for CaseExpr {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        fmt.subregion().clear_trailing_columns(true).enter(|fmt| {
-            self.case.format(fmt)?;
-            fmt.write_space()?;
+    fn format(&self, fmt: &mut Formatter) {
+        self.case.format(fmt);
+        fmt.add_space();
+        self.value.format(fmt);
+        fmt.add_space();
+        self.of.format(fmt);
+        self.clauses.format(fmt);
+        self.end.format(fmt);
+    }
+}
 
-            fmt.subregion()
-                .reset_trailing_columns(3) // " of"
-                .enter(|fmt| self.value.format(fmt))?;
-            fmt.write_space()?;
+#[derive(Debug, Clone, Span, Parse)]
+struct End(EndKeyword);
 
-            self.of.format(fmt)?;
-            self.clauses.format(fmt)?;
-            self.end.format(fmt)?;
-
-            Ok(())
-        })
+impl Format for End {
+    fn format(&self, fmt: &mut Formatter) {
+        fmt.subregion(Indent::inherit(), Newline::Always, |fmt| self.0.format(fmt));
     }
 }
 
@@ -63,25 +64,13 @@ struct CaseClause {
 
 /// `if` (`$CLAUSE` `;`?)+ `end`
 ///
-/// - $CLAUSE: `$GUARD` `->` `$BODY`
+/// - $CLAUSE: `$GUARD` `->` [Body]
 /// - $GUARD: ([Expr] (`,` | `;`)?)+
-/// - $BODY: ([Expr] `,`?)+
-#[derive(Debug, Clone, Span, Parse)]
+#[derive(Debug, Clone, Span, Parse, Format)]
 pub struct IfExpr {
     r#if: IfKeyword,
     clauses: Block<Clauses<IfClause>>,
-    end: EndKeyword,
-}
-
-impl Format for IfExpr {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        fmt.subregion().clear_trailing_columns(true).enter(|fmt| {
-            self.r#if.format(fmt)?;
-            self.clauses.format(fmt)?;
-            self.end.format(fmt)?;
-            Ok(())
-        })
-    }
+    end: End,
 }
 
 #[derive(Debug, Clone, Span, Parse, Format)]
@@ -93,67 +82,61 @@ struct IfClause {
 #[derive(Debug, Clone, Span, Parse, Format)]
 struct GuardCondition(NonEmptyItems<Expr, Either<CommaSymbol, SemicolonSymbol>>);
 
-/// `begin` `$BODY` `end`
+/// `begin` [Body] `end`
 ///
-/// - $BODY: ([Expr] `,`?)+
 #[derive(Debug, Clone, Span, Parse, Format)]
 pub struct BeginExpr {
     begin: BeginKeyword,
-    exprs: Block<NonEmptyItems<Expr>>,
-    end: EndKeyword,
+    exprs: Body,
+    end: End,
 }
 
 /// `receive` (`$CLAUSE` `;`?)* `$TIMEOUT`? `end`
 ///
-/// - $CLAUSE: `$PATTERN` (`when` `$GUARD`)? `->` `$BODY`
+/// - $CLAUSE: `$PATTERN` (`when` `$GUARD`)? `->` [Body]
 /// - $PATTERN: [Expr]
 /// - $GUARD: ([Expr] (`,` | `;`)?)+
-/// - $TIMEOUT: `after` [Expr] `->` `$BODY`
-/// - $BODY: ([Expr] `,`?)+
-#[derive(Debug, Clone, Span, Parse)]
+/// - $TIMEOUT: `after` [Expr] `->` [Body]
+#[derive(Debug, Clone, Span, Parse, Format)]
 pub struct ReceiveExpr {
     receive: ReceiveKeyword,
     clauses: Block<Maybe<Clauses<CaseClause>>>,
     timeout: Maybe<ReceiveTimeout>,
-    end: EndKeyword,
+    end: End,
 }
 
-impl Format for ReceiveExpr {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        fmt.subregion().clear_trailing_columns(true).enter(|fmt| {
-            self.receive.format(fmt)?;
-            self.clauses.format(fmt)?;
-            self.timeout.format(fmt)?;
-            self.end.format(fmt)?;
-            Ok(())
-        })
-    }
-}
-
-#[derive(Debug, Clone, Span, Parse, Format)]
+#[derive(Debug, Clone, Span, Parse)]
 struct ReceiveTimeout {
     after: AfterKeyword,
     clause: Block<ReceiveTimeoutClause>,
 }
 
+impl Format for ReceiveTimeout {
+    fn format(&self, fmt: &mut Formatter) {
+        fmt.subregion(Indent::inherit(), Newline::Always, |fmt| {
+            self.after.format(fmt);
+            self.clause.format(fmt);
+        });
+    }
+}
+
 #[derive(Debug, Clone, Span, Parse, Format)]
 struct ReceiveTimeoutClause {
     timeout: WithArrow<Expr>,
     body: Body,
 }
 
-/// `try` `$BODY` `$BRANCHES`? `$CATCH`? `$AFTER`? `end`
+/// `try` [Body] `$BRANCHES`? `$CATCH`? `$AFTER`? `end`
 ///
-/// - $BODY: ([Expr] `,`?)+
 /// - $BRANCHES: (`of` (`$BRANCH_CLAUSE` `;`?)+)
-/// - $BRANCH_CLAUSE: `$PATTERN` (`when` `$GUARD`)? `->` `$BODY`
+/// - $BRANCH_CLAUSE: `$PATTERN` (`when` `$GUARD`)? `->` [Body]
 /// - $PATTERN: [Expr]
 /// - $GUARD: ([Expr] (`,` | `;`)?)+
 /// - $CATCH: `catch` (`$CATCH_CLAUSE` `;`?)+
-/// - $CATCH_CLAUSE: `$ERROR_CLASS`? [Expr] `$STACKTRACE`? (`when` `$GUARD`)? `->` `$BODY`
+/// - $CATCH_CLAUSE: `$ERROR_CLASS`? [Expr] `$STACKTRACE`? (`when` `$GUARD`)? `->` [Body]
 /// - $ERROR_CLASS: ([AtomToken] | [VariableToken]) `:`
 /// - $STACKTRACE: `:` [VariableToken]
-/// - $AFTER: `after` `$BODY`
+/// - $AFTER: `after` [Body]
 #[derive(Debug, Clone, Span, Parse)]
 pub struct TryExpr {
     r#try: TryKeyword,
@@ -161,22 +144,22 @@ pub struct TryExpr {
     clauses: Maybe<(OfKeyword, Block<Clauses<CaseClause>>)>,
     catch: Maybe<TryCatch>,
     after: Maybe<TryAfter>,
-    end: EndKeyword,
+    end: End,
 }
 
 impl Format for TryExpr {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        fmt.subregion().clear_trailing_columns(true).enter(|fmt| {
-            self.r#try.format(fmt)?;
-            self.body.format(fmt)?;
-            fmt.write_newline()?;
-            self.clauses.format(fmt)?;
-            self.catch.format(fmt)?;
-            self.after.format(fmt)?;
-            fmt.write_newline()?;
-            self.end.format(fmt)?;
-            Ok(())
-        })
+    fn format(&self, fmt: &mut Formatter) {
+        self.r#try.format(fmt);
+        self.body.format(fmt);
+        fmt.add_newline();
+        self.clauses.format(fmt);
+        fmt.subregion(Indent::inherit(), Newline::Always, |fmt| {
+            self.catch.format(fmt)
+        });
+        fmt.subregion(Indent::inherit(), Newline::Always, |fmt| {
+            self.after.format(fmt)
+        });
+        self.end.format(fmt);
     }
 }
 
@@ -213,13 +196,12 @@ pub struct CatchExpr {
 }
 
 impl Format for CatchExpr {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        self.catch.format(fmt)?;
-        fmt.write_space()?;
-        fmt.subregion()
-            .current_column_as_indent()
-            .enter(|fmt| self.expr.format(fmt))?;
-        Ok(())
+    fn format(&self, fmt: &mut Formatter) {
+        self.catch.format(fmt);
+        fmt.add_space();
+        fmt.subregion(Indent::CurrentColumn, Newline::Never, |fmt| {
+            self.expr.format(fmt)
+        });
     }
 }
 
@@ -227,22 +209,17 @@ impl Format for CatchExpr {
 struct Block<T>(T);
 
 impl<T: Format> Format for Block<T> {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        fmt.subregion()
-            .clear_trailing_columns(true)
-            .indent_offset(4)
-            .enter(|fmt| {
-                fmt.write_newline()?;
-                self.0.format(fmt)?;
-                fmt.write_newline()?;
-                Ok(())
-            })
+    fn format(&self, fmt: &mut Formatter) {
+        fmt.subregion(Indent::Offset(4), Newline::Always, |fmt| {
+            self.0.format(fmt);
+            fmt.add_newline();
+        });
     }
 }
 
 #[cfg(test)]
 mod tests {
-    use super::*;
+    use crate::items::Expr;
 
     #[test]
     fn case_works() {
@@ -294,33 +271,37 @@ mod tests {
     fn receive_works() {
         let texts = [
             indoc::indoc! {"
-                receive
-                    {A, B} ->
-                        [A, B];
-                    A when is_integer(A);
-                           is_atom(A) ->
-                        A
-                end"},
+            %---10---|%---20---|
+            receive
+                {A, B} ->
+                    [A, B];
+                A when is_integer(A);
+                       is_atom(A) ->
+                    A
+            end"},
             indoc::indoc! {"
-                receive
-                    A ->
-                        A
-                after
-                    1000 ->
-                        timeout
-                end"},
+            %---10---|%---20---|
+            receive
+                A ->
+                    A
+            after
+                1000 ->
+                    timeout
+            end"},
             indoc::indoc! {"
-                receive
-                after
-                    N ->
-                        timeout
-                end"},
+            %---10---|%---20---|
+            receive
+            after
+                N ->
+                    timeout
+            end"},
             indoc::indoc! {"
-                receive
-                after
-                    infinity ->
-                        timeout
-                end"},
+            %---10---|%---20---|
+            receive
+            after
+                infinity ->
+                    timeout
+            end"},
         ];
         for text in texts {
             crate::assert_format!(text, Expr);
@@ -350,51 +331,56 @@ mod tests {
     fn try_works() {
         let texts = [
             indoc::indoc! {"
-                try
-                    1
-                after
+            %---10---|%---20---|
+            try
+                1
+            after
+                2
+            end"},
+            indoc::indoc! {"
+            %---10---|%---20---|
+            try
+                1,
+                2,
+                3
+            catch
+                E ->
+                    E
+            end"},
+            indoc::indoc! {"
+            %---10---|%---20---|
+            try
+                X
+            of
+                {_, _} ->
+                    1;
+                [_, _] ->
                     2
-                end"},
+            catch
+                _:E:Stacktrace
+                  when is_atom(E) ->
+                    foo
+            after
+                bar
+            end"},
             indoc::indoc! {"
-                try
-                    1,
-                    2,
-                    3
-                catch
-                    E ->
-                        E
-                end"},
+            %---10---|%---20---|
+            try
+                foo()
+            catch
+                _:#foo{} = E ->
+                    E
+            end"},
             indoc::indoc! {"
-                try
-                    X
-                of
-                    {_, _} ->
-                        1;
-                    [_, _] ->
-                        2
-                catch
-                    _:E:Stacktrace
-                      when is_atom(E) ->
-                        foo
-                after
-                    bar
-                end"},
-            indoc::indoc! {"
-                try
-                    foo()
-                catch
-                    _:#foo{} = E ->
-                        E
-                end"},
-            indoc::indoc! {"
-                try
-                    foo()
-                catch
-                    throw:_ ->
-                        throw;
-                    error:_ ->
-                        error
-                end"},
+            %---10---|%---20---|
+            try
+                foo()
+            catch
+                throw:_ ->
+                    throw;
+                error:_ ->
+                    error
+            end"},
         ];
         for text in texts {
             crate::assert_format!(text, Expr);
