@@ -1,9 +1,12 @@
-use crate::format::{self, Format};
-use crate::items::expressions::{AtomLikeExpr, Body, Expr, IntegerLikeExpr};
-use crate::items::generics::{Clauses, Maybe, Null, Params, WithArrow, WithGuard};
+use crate::format::{Format, Formatter, Indent, Newline};
+use crate::items::components::{Clauses, Maybe, Null};
+use crate::items::expressions::components::FunctionClause;
+use crate::items::expressions::BaseExpr;
 use crate::items::keywords::{EndKeyword, FunKeyword};
 use crate::items::symbols::{ColonSymbol, SlashSymbol};
 use crate::items::tokens::VariableToken;
+#[cfg(doc)]
+use crate::items::Expr;
 use crate::parse::Parse;
 use crate::span::Span;
 
@@ -22,10 +25,10 @@ pub enum FunctionExpr {
 #[derive(Debug, Clone, Span, Parse, Format)]
 pub struct DefinedFunctionExpr {
     fun: Fun,
-    module: Maybe<(AtomLikeExpr, ColonSymbol)>,
-    name: AtomLikeExpr,
+    module: Maybe<(BaseExpr, ColonSymbol)>,
+    name: BaseExpr,
     slash: SlashSymbol,
-    arity: IntegerLikeExpr,
+    arity: BaseExpr,
 }
 
 /// `fun` (`$CLAUSE` `;`?)+ `end`
@@ -36,8 +39,7 @@ pub struct DefinedFunctionExpr {
 #[derive(Debug, Clone, Span, Parse, Format)]
 pub struct AnonymousFunctionExpr {
     fun: Fun,
-    clauses: FunctionClauses<Null>,
-    end: EndKeyword,
+    clauses_and_end: FunctionClausesAndEnd<Null>,
 }
 
 /// `fun` (`$CLAUSE` `;`?)+ `end`
@@ -48,66 +50,48 @@ pub struct AnonymousFunctionExpr {
 #[derive(Debug, Clone, Span, Parse, Format)]
 pub struct NamedFunctionExpr {
     fun: Fun,
-    clauses: FunctionClauses<VariableToken>,
-    end: EndKeyword,
+    clauses_and_end: FunctionClausesAndEnd<VariableToken>,
 }
 
 #[derive(Debug, Clone, Span, Parse)]
 struct Fun(FunKeyword);
 
 impl Format for Fun {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        self.0.format(fmt)?;
-        fmt.write_space()?;
-        Ok(())
+    fn format(&self, fmt: &mut Formatter) {
+        self.0.format(fmt);
+        fmt.add_space();
     }
 }
 
 #[derive(Debug, Clone, Span, Parse)]
-struct FunctionClauses<Name>(Clauses<FunctionClause<Name>>);
-
-impl<Name: Format> Format for FunctionClauses<Name> {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        if self.0.items().len() == 1 && self.0.items()[0].body.exprs().len() == 1 {
-            let ok = fmt
-                .subregion()
-                .forbid_multi_line()
-                .forbid_too_long_line()
-                .trailing_columns(4) // "end"
-                .check_trailing_columns(true)
-                .enter(|fmt| {
-                    let clause = &self.0.items()[0];
-                    clause.name.format(fmt)?;
-                    clause.params.format(fmt)?;
-                    fmt.write_space()?;
-                    clause.body.exprs()[0].format(fmt)?;
-                    fmt.write_space()?;
-                    Ok(())
-                })
-                .is_ok();
-            if ok {
-                return Ok(());
-            }
-        }
-
-        fmt.subregion().current_column_as_indent().enter(|fmt| {
-            self.0.format(fmt)?;
-            fmt.write_newline()?;
-            Ok(())
-        })
-    }
+struct FunctionClausesAndEnd<Name> {
+    clauses: Clauses<FunctionClause<Name>>,
+    end: EndKeyword,
 }
 
-#[derive(Debug, Clone, Span, Parse, Format)]
-pub(crate) struct FunctionClause<Name> {
-    name: Name,
-    params: WithArrow<WithGuard<Params<Expr>, Expr>>,
-    body: Body,
+impl<Name: Format> Format for FunctionClausesAndEnd<Name> {
+    fn format(&self, fmt: &mut Formatter) {
+        if self.clauses.items().len() == 1 && self.clauses.items()[0].body().exprs().len() == 1 {
+            let clause = &self.clauses.items()[0];
+            fmt.subregion(Indent::CurrentColumn, Newline::Never, |fmt| {
+                clause.format_maybe_one_line_body(fmt);
+                fmt.subregion(
+                    Indent::ParentOffset(0),
+                    Newline::IfTooLongOrMultiLineParent,
+                    |fmt| self.end.format(fmt),
+                );
+            });
+        } else {
+            self.clauses.format(fmt);
+            fmt.add_newline();
+            self.end.format(fmt);
+        }
+    }
 }
 
 #[cfg(test)]
 mod tests {
-    use super::*;
+    use crate::items::expressions::Expr;
 
     #[test]
     fn defined_function_works() {
@@ -165,6 +149,11 @@ mod tests {
             "fun Foo() -> hi end",
             indoc::indoc! {"
             %---10---|%---20---|
+            fun Foo() ->
+                    hello
+            end"},
+            indoc::indoc! {"
+            %---10---|%---20---|
             fun Foo(a) ->
                     a;
                 Foo(A) ->
