@@ -1,9 +1,10 @@
-use crate::format::{self, Format};
-use crate::items::expressions::{Either, Expr};
-use crate::items::generics::{BinaryOpLike, BinaryOpStyle, TupleLike};
+use crate::format::{Format, Indent, Newline};
+use crate::items::components::{BinaryOpLike, BinaryOpStyle, Element, TupleLike};
+use crate::items::expressions::Either;
 use crate::items::symbols::{DotSymbol, MatchSymbol, SharpSymbol};
 use crate::items::tokens::AtomToken;
 use crate::items::variables::UnderscoreVariable;
+use crate::items::Expr;
 use crate::parse::{self, Parse, ResumeParse};
 use crate::span::Span;
 
@@ -45,7 +46,7 @@ pub struct RecordConstructExpr {
 /// - $VALUE: [Expr]
 /// - $NAME: [AtomToken]
 /// - $FIELD: [AtomToken]
-#[derive(Debug, Clone, Span, Parse)]
+#[derive(Debug, Clone, Span, Parse, Format)]
 pub struct RecordAccessExpr {
     value: Expr,
     index: RecordIndexExpr,
@@ -60,17 +61,6 @@ impl ResumeParse<Expr> for RecordAccessExpr {
     }
 }
 
-impl Format for RecordAccessExpr {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        self.value.format(fmt)?;
-        if self.value.is_integer_token() {
-            fmt.write_space()?;
-        }
-        self.index.format(fmt)?;
-        Ok(())
-    }
-}
-
 /// `#` `$NAME` `.` `$FIELD`
 ///
 /// - $NAME: [AtomToken]
@@ -88,7 +78,7 @@ pub struct RecordIndexExpr {
 /// - $VALUE: [Expr]
 /// - $NAME: [AtomToken]
 /// - $FIELD: [AtomToken]
-#[derive(Debug, Clone, Span, Parse)]
+#[derive(Debug, Clone, Span, Parse, Format)]
 pub struct RecordUpdateExpr {
     value: Expr,
     sharp: SharpSymbol,
@@ -107,36 +97,19 @@ impl ResumeParse<Expr> for RecordUpdateExpr {
     }
 }
 
-impl Format for RecordUpdateExpr {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        self.value.format(fmt)?;
-        if self.value.is_integer_token() {
-            fmt.write_space()?;
-        }
-        self.sharp.format(fmt)?;
-        self.name.format(fmt)?;
-        self.fields.format(fmt)?;
-        Ok(())
-    }
-}
-
-#[derive(Debug, Clone, Span, Parse, Format)]
+#[derive(Debug, Clone, Span, Parse, Format, Element)]
 struct RecordField(BinaryOpLike<Either<AtomToken, UnderscoreVariable>, RecordFieldDelimiter, Expr>);
 
 #[derive(Debug, Clone, Span, Parse, Format)]
 struct RecordFieldDelimiter(MatchSymbol);
 
 impl BinaryOpStyle for RecordFieldDelimiter {
-    fn indent_offset(&self) -> usize {
-        4
+    fn indent(&self) -> Indent {
+        Indent::Offset(4)
     }
 
-    fn allow_newline(&self) -> bool {
-        true
-    }
-
-    fn should_pack(&self) -> bool {
-        false
+    fn newline(&self) -> Newline {
+        Newline::IfTooLongOrMultiLine
     }
 }
 
@@ -188,7 +161,10 @@ mod tests {
         let texts = [
             "M#foo{}",
             "88 #foo{}",
-            "M#baz{qux = 1}#foo.bar",
+            indoc::indoc! {"
+            %---10---|%---20---|
+            M#baz{qux =
+                      1}#foo.bar"},
             indoc::indoc! {"
             %---10---|%---20---|
             M#foo.bar#baz{qux =
