@@ -1,40 +1,23 @@
 //! Erlang types.
 //!
 //! <https://www.erlang.org/doc/reference_manual/typespec.html>
-use crate::format::{self, Format};
-use crate::items::generics::{
-    Args, BinaryOpLike, BinaryOpStyle, BitstringLike, Either, ListLike, Maybe, NonEmptyItems,
-    Params, Parenthesized, TupleLike, UnaryOpLike,
-};
-use crate::items::keywords::{
-    BandKeyword, BnotKeyword, BorKeyword, BslKeyword, BsrKeyword, BxorKeyword, DivKeyword,
-    FunKeyword, RemKeyword,
+use self::components::{BinaryOp, BitstringItem, UnaryOp};
+use crate::format::{Format, Formatter, Indent, Newline};
+use crate::items::components::{
+    Args, BinaryOpLike, BinaryOpStyle, BitstringLike, Either, Element, ListLike, MapLike, Maybe,
+    NonEmptyItems, Params, Parenthesized, TupleLike, UnaryOpLike,
 };
+use crate::items::keywords::FunKeyword;
 use crate::items::symbols::{
-    ColonSymbol, DoubleColonSymbol, DoubleDotSymbol, HyphenSymbol, MultiplySymbol, PlusSymbol,
-    RightArrowSymbol, SharpSymbol, TripleDotSymbol, VerticalBarSymbol,
+    ColonSymbol, DoubleColonSymbol, RightArrowSymbol, SharpSymbol, TripleDotSymbol,
+    VerticalBarSymbol,
 };
 use crate::items::tokens::{AtomToken, CharToken, IntegerToken, VariableToken};
 use crate::items::Type;
 use crate::parse::{self, Parse, ResumeParse};
 use crate::span::Span;
 
-#[derive(Debug, Clone, Span, Format)]
-enum NonUnionType {
-    BinaryOp(Box<BinaryOpType>),
-    NonLeftRecursive(NonLeftRecursiveType),
-}
-
-impl Parse for NonUnionType {
-    fn parse(ts: &mut parse::TokenStream) -> parse::Result<Self> {
-        let expr: NonLeftRecursiveType = ts.parse()?;
-        if ts.peek::<BinaryOp>().is_some() {
-            ts.resume_parse(expr).map(Self::BinaryOp)
-        } else {
-            Ok(Self::NonLeftRecursive(expr))
-        }
-    }
-}
+pub mod components;
 
 /// [Type] `|` [Type]
 #[derive(Debug, Clone, Span, Parse, Format)]
@@ -44,15 +27,33 @@ pub struct UnionType(NonEmptyItems<NonUnionType, UnionDelimiter>);
 struct UnionDelimiter(VerticalBarSymbol);
 
 impl Format for UnionDelimiter {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        fmt.write_space()?;
-        self.0.format(fmt)?;
-        Ok(())
+    fn format(&self, fmt: &mut Formatter) {
+        fmt.add_space();
+        self.0.format(fmt);
+        fmt.add_space();
     }
 }
 
+#[derive(Debug, Clone, Span, Format)]
+enum NonUnionType {
+    Base(BaseType),
+    BinaryOp(Box<BinaryOpType>),
+}
+
+impl Parse for NonUnionType {
+    fn parse(ts: &mut parse::TokenStream) -> parse::Result<Self> {
+        let expr: BaseType = ts.parse()?;
+        if ts.peek::<BinaryOp>().is_some() {
+            ts.resume_parse(expr).map(Self::BinaryOp)
+        } else {
+            Ok(Self::Base(expr))
+        }
+    }
+}
+
+// Non left-recursive type.
 #[derive(Debug, Clone, Span, Parse, Format)]
-enum NonLeftRecursiveType {
+enum BaseType {
     Mfargs(Box<MfargsType>),
     List(Box<ListType>),
     Tuple(Box<TupleType>),
@@ -75,74 +76,28 @@ pub struct AnnotatedVariableType {
 }
 
 impl Format for AnnotatedVariableType {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        self.variable.format(fmt)?;
-        fmt.write_space()?;
-        self.colon.format(fmt)?;
-        fmt.write_space()?;
-        self.ty.format(fmt)?;
-        Ok(())
+    fn format(&self, fmt: &mut Formatter) {
+        self.variable.format(fmt);
+        fmt.add_space();
+        self.colon.format(fmt);
+        fmt.add_space();
+        self.ty.format(fmt);
     }
 }
 
-/// [Type] `$OP` [Type]
-///
-/// - $OP: [BinaryOp]
+/// [Type] [BinaryOp] [Type]
 #[derive(Debug, Clone, Span, Parse, Format)]
-pub struct BinaryOpType(BinaryOpLike<NonLeftRecursiveType, BinaryOp, Type>);
+pub struct BinaryOpType(BinaryOpLike<BaseType, BinaryOp, Type>);
 
-impl ResumeParse<NonLeftRecursiveType> for BinaryOpType {
-    fn resume_parse(
-        ts: &mut parse::TokenStream,
-        left: NonLeftRecursiveType,
-    ) -> parse::Result<Self> {
+impl ResumeParse<BaseType> for BinaryOpType {
+    fn resume_parse(ts: &mut parse::TokenStream, left: BaseType) -> parse::Result<Self> {
         ts.resume_parse(left).map(Self)
     }
 }
 
-/// `*` | `+` | `-` | `div` | `rem` | `band` | `bor` | `bxor` | `bsl` | `bsr` | `..`
+/// [UnaryOp] [Type]
 #[derive(Debug, Clone, Span, Parse, Format)]
-pub enum BinaryOp {
-    Mul(MultiplySymbol),
-    Plus(PlusSymbol),
-    Minus(HyphenSymbol),
-    Div(DivKeyword),
-    Rem(RemKeyword),
-    Band(BandKeyword),
-    Bor(BorKeyword),
-    Bxor(BxorKeyword),
-    Bsl(BslKeyword),
-    Bsr(BsrKeyword),
-    Range(DoubleDotSymbol),
-}
-
-impl BinaryOpStyle for BinaryOp {
-    fn indent_offset(&self) -> usize {
-        0
-    }
-
-    fn allow_newline(&self) -> bool {
-        true
-    }
-
-    fn should_pack(&self) -> bool {
-        true
-    }
-}
-
-/// `$OP` [Type]
-///
-/// - $OP: [UnaryOp]
-#[derive(Debug, Clone, Span, Parse, Format)]
-pub struct UnaryOpType(UnaryOpLike<UnaryOp, NonLeftRecursiveType>);
-
-/// `+` | `-` | `bnot`
-#[derive(Debug, Clone, Span, Parse, Format)]
-pub enum UnaryOp {
-    Plus(PlusSymbol),
-    Minus(HyphenSymbol),
-    Bnot(BnotKeyword),
-}
+pub struct UnaryOpType(UnaryOpLike<UnaryOp, BaseType>);
 
 /// `fun` `(` (`$PARAMS` `->` `$RETURN`)? `)`
 ///
@@ -155,11 +110,10 @@ pub struct FunctionType {
 }
 
 impl Format for FunctionType {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
-        self.fun.format(fmt)?;
-        fmt.write_space()?;
-        self.params_and_return.format(fmt)?;
-        Ok(())
+    fn format(&self, fmt: &mut Formatter) {
+        self.fun.format(fmt);
+        fmt.add_space();
+        self.params_and_return.format(fmt);
     }
 }
 
@@ -169,16 +123,12 @@ type FunctionParamsAndReturn = BinaryOpLike<FunctionParams, RightArrowDelimiter,
 struct RightArrowDelimiter(RightArrowSymbol);
 
 impl BinaryOpStyle for RightArrowDelimiter {
-    fn indent_offset(&self) -> usize {
-        8
+    fn indent(&self) -> Indent {
+        Indent::Offset(8)
     }
 
-    fn allow_newline(&self) -> bool {
-        true
-    }
-
-    fn should_pack(&self) -> bool {
-        false
+    fn newline(&self) -> Newline {
+        Newline::IfTooLongOrMultiLine
     }
 }
 
@@ -213,20 +163,27 @@ pub struct MfargsType {
 ///
 /// - $ITEM: [Type] | `...`
 #[derive(Debug, Clone, Span, Parse, Format)]
-pub struct ListType(ListLike<Either<Type, TripleDotSymbol>>);
+pub struct ListType(ListLike<ListItem>);
+
+#[derive(Debug, Clone, Span, Parse, Format)]
+struct ListItem(Either<Type, TripleDotSymbol>);
+
+impl Element for ListItem {
+    fn is_packable(&self) -> bool {
+        false
+    }
+}
 
 /// `{` ([Type] `,`)* `}`
 #[derive(Debug, Clone, Span, Parse, Format)]
-pub struct TupleType(TupleLike<Type>);
+pub struct TupleType(TupleLike<TupleItem>);
+
+#[derive(Debug, Clone, Span, Parse, Format, Element)]
+struct TupleItem(Type);
 
 /// `#` `{` ([Type] (`:=` | `=>`) [Type] `,`?)* `}`
 #[derive(Debug, Clone, Span, Parse, Format)]
-pub struct MapType {
-    sharp: SharpSymbol,
-    items: TupleLike<MapItem>,
-}
-
-type MapItem = BinaryOpLike<Type, crate::items::expressions::maps::MapDelimiter, Type>;
+pub struct MapType(MapLike<Type>);
 
 /// `#` `$NAME` `{` (`$FIELD` `,`?)* `}`
 ///
@@ -239,22 +196,19 @@ pub struct RecordType {
     fields: TupleLike<RecordItem>,
 }
 
-type RecordItem = BinaryOpLike<AtomToken, DoubleColonDelimiter, Type>;
+#[derive(Debug, Clone, Span, Parse, Format, Element)]
+struct RecordItem(BinaryOpLike<AtomToken, DoubleColonDelimiter, Type>);
 
 #[derive(Debug, Clone, Span, Parse, Format)]
 struct DoubleColonDelimiter(DoubleColonSymbol);
 
 impl BinaryOpStyle for DoubleColonDelimiter {
-    fn indent_offset(&self) -> usize {
-        4
+    fn indent(&self) -> Indent {
+        Indent::Offset(4)
     }
 
-    fn allow_newline(&self) -> bool {
-        true
-    }
-
-    fn should_pack(&self) -> bool {
-        false
+    fn newline(&self) -> Newline {
+        Newline::IfTooLongOrMultiLine
     }
 }
 
@@ -263,23 +217,7 @@ impl BinaryOpStyle for DoubleColonDelimiter {
 /// - $BITS_SIZE: `_` `:` [Type]
 /// - $UNIT_SIZE: `_` `:` `_` `*` [Type]
 #[derive(Debug, Clone, Span, Parse, Format)]
-pub struct BitstringType(BitstringLike<Either<BitstringUnitSize, BitstringBitsSize>>);
-
-#[derive(Debug, Clone, Span, Parse, Format)]
-struct BitstringBitsSize {
-    underscore: VariableToken,
-    colon: ColonSymbol,
-    size: Type,
-}
-
-#[derive(Debug, Clone, Span, Parse, Format)]
-struct BitstringUnitSize {
-    underscore0: VariableToken,
-    colon: ColonSymbol,
-    underscore1: VariableToken,
-    mul: MultiplySymbol,
-    size: Type,
-}
+pub struct BitstringType(BitstringLike<BitstringItem>);
 
 #[cfg(test)]
 mod tests {
@@ -291,9 +229,15 @@ mod tests {
             "foo()",
             "foo:bar(A, 1)",
             indoc::indoc! {"
+            %---10---|%---20---|
             foo:bar(A,
-                    1,
-                    baz(),
+                    BB,
+                    baz())"},
+            indoc::indoc! {"
+            %---10---|%---20---|
+            foo:bar(A,
+                    B,
+                    baz(12, 34),
                     qux())"},
         ];
         for text in texts {
@@ -324,9 +268,15 @@ mod tests {
             "{foo()}",
             "{atom, 1}",
             indoc::indoc! {"
-                {foo(),
-                 bar(),
-                 [baz()]}"},
+            %---10---|%---20---|
+            {foo(),
+             bar(),
+             [baz()]}"},
+            indoc::indoc! {"
+            %---10---|%---20---|
+            {foo(),
+             bar(),
+             [baz(A, B)]}"},
         ];
         for text in texts {
             crate::assert_format!(text, Type);
@@ -349,6 +299,7 @@ mod tests {
                    bbb,
                    ccc}}"},
             indoc::indoc! {"
+            %---10---|%---20---|
             #{a => b,
               1 := 2,
               atom() := atom()}"},
