@@ -11,5 +11,7 @@
 
 hello(A) ->
     io?BAR "hello ~p\n", [A]).
-main() ?a(?qux a, b), c],
-                    [1, hello(world?quux 3].
+main() ?a(?qux a, b)
+    , c],
+    [1, hello(world?quux
+    3].
